// KolACMWavePlayer (c) 2005, Thaddy de Koning
// Plays wave files and also MP3 files through the ACM codec if installed
// Original Copyright(c) 2001-2003 by Ing. Tomas Koutny.
// (rawos@rawos.com, xdtktom@quick.cz)
//  Based on Version 2.11

//
// Remarks
// =======
//
// * Some audio codecs allow only one decoding process per application.
// * To play MPEG audio you need to have already installed appropriate codec,
//


unit KolACMWavePlayer;

interface

uses Kol, err, Windows, MMSystem, Messages, Objects;


type
  TWavePlayingState = (wpStopped, wpPlaying, wpPaused);

  PWavePlayer = ^TWavePlayer;
  TWavePlayer = object(TObj)
  private
    FDevice:HWAVEOUT;                //Device to play the wave
    FBuffers:array[0..7] of TWAVEHDR;//8 buffers to have not chopping playback
    FBufferSize:integer;             //How much memory to allocate per buffer
    FRepeat:boolean;                 //Whether to play in a loop
    FDataStart:integer;              //Where the headers end
    FDataEnd:integer;                //Length of data (+1 to avoid one asm add
                                     //in LoadBuffer) ; data chunk is not
                                     //always the last one
    FPlayingBufferCount:integer;     //Number of buffers being played
    procedure LoadBuffer(AIndex:integer);  //Loads values into selected buffer
    function LoadWaveFile:PWaveFormatEx;
    function LoadMPEG1File:PWaveFormatEx;
  protected
    FState:TWavePlayingState;  //Internal (and desired) state
    FSource:PStream;           //Source the wave will be played from
    //Following two variables may be figured out by setting FRepeat to true
    FLoopsToPlay:integer;      //Remaining count of loops to be played
    FDesiredLoops:integer;     //Desired count of loops to be played
    FBeepAtLeast:boolean;
    FOnFinishedPlayback:TOnEvent;
    FWnd:HWND;
    procedure MsgHandler(var msg:TMessage);
  public
    procedure init; virtual;
    destructor Destroy; virtual;
    //All functions below return true if they proceed successfuly
    //and false if not.
    function Play:boolean;  virtual;     //Starts or unpauses the wave playing
    function Stop:boolean;  virtual;    //Stops the current wave playing
    function Reset:boolean; virtual;    //Rewinds to zero and stops playing
    function Pause:boolean; virtual;    //Pauses or unpauses the wave playing
    property State:TWavePlayingState read FState;
    //The Source property MUST be set before the wave playing is started!
    //(The riff file format uses 32 bit adressing => the standard stream
    // is sufficient.)
    property Source:PStream read FSource write FSource;
     //The Loop property should be also set before the wave playing is started,
    //because all internal buffers are being queued regardless of this value.
    property Loop:boolean read FRepeat write FRepeat;
    property Loops:integer read FDesiredLoops write FDesiredLoops;
    property BeepAtLeast:boolean read FBeepAtLeast write FBeepAtLeast;
    property OnFinishedPlayback:TOnEvent read FOnFinishedPlayback write FOnFinishedPlayback;
  end;


implementation

const

  WaveHdrSize = sizeof(TWAVEHDR); //Why to call sizeof so much times?

//Declare the wave file header's structure
type
  TWAVHeader = packed record
    RIFFHeader: packed array [0..3] of Char;   //'RIFF'
    FileSize: Longint;
    WAVEHeader: packed array [0..3] of Char;   //'WAVE'
    FormatHeader: packed array [0..3] of Char; //'fmt ' - there's a space!
    FormatSize: Longint;
    FormatEx:TWaveFormatEx;
  end;

  TChunk = packed record
    Signature:LongWord;
    Length:longint;
  end;

  //Since the support of MPEG1
  //Following is taken from MSDN Library
  PMPEG1WAVEFORMAT = ^TMPEG1WAVEFORMAT;
  TMPEG1WAVEFORMAT = packed record
    wfx:TWaveFormatEx;
    fwHeadLayer:word;
    dwHeadBitrate:DWORD;
    fwHeadMode:WORD;
    fwHeadModeExt:WORD;
    wHeadEmphasis:WORD;
    fwHeadFlags:WORD;
    dwPTSLow:DWORD;
    dwPTSHigh:DWORD;
  end;

  PMPEGLAYER3WAVEFORMAT = ^TMPEGLAYER3WAVEFORMAT;
  TMPEGLAYER3WAVEFORMAT = packed record
    wfx:tWAVEFORMATEX;
    wID:WORD;
    fdwFlags:DWORD;
    nBlockSize:WORD;
    nFramesPerBlock:WORD;
    nCodecDelay:WORD;
  end;


const
  WAVE_FORMAT_MPEG = $50;
  WAVE_FORMAT_MPEGLAYER3 = $55;

  ACM_MPEG_LAYER1  = 1;
  ACM_MPEG_LAYER2  = 2;
  ACM_MPEG_LAYER3  = 4;

  ACM_MPEG_STEREO        = 1;
  ACM_MPEG_JOINTSTEREO   = 2;
  ACM_MPEG_DUALCHANNEL   = 4;
  ACM_MPEG_SINGLECHANNEL = 8;

  ACM_MPEG_ID_MPEG1 = $10;
  MPEGLAYER3_ID_MPEG = 1;

  MPEGLAYER3_FLAG_PADDING_ON  = 1;
  MPEGLAYER3_FLAG_PADDING_OFF = 2;




type
  EWPlayer = Exception;


procedure TWavePlayer.init;
begin
  inherited;
  FState:=wpStopped;
  FRepeat:=false;
  BeepAtLeast:=false;
  FWnd:=AllocateHWnd(MsgHandler);
end;

destructor TWavePlayer.Destroy;
begin
  try
    Stop;
  finally
    DeallocateHWnd(FWnd);
  end;
  inherited;
end;

procedure TWavePlayer.MsgHandler(var msg:TMessage);
var i:integer;
begin
  with Msg do
    case Msg of
      MM_WOM_DONE:begin
                    //Using hwo is faster than FDevice since it is in a register
                    waveOutUnprepareHeader(wParam,
                                           PWAVEHDR(lParam),
                                           WaveHdrSize);
                    dec(FPlayingBufferCount);

                    if FState = wpStopped then exit;
                       //Presume that stopping is in progress

                    //Get data
                    LoadBuffer(PWAVEHDR(lParam)^.dwUser);

                    //Check for end of wave
                    if PWAVEHDR(lParam)^.dwBufferLength = 0 then
                      begin
                        if FRepeat or (FLoopsToPlay>0) then  //Play once again or not?
                          begin
                            FSource.Position:=FDataStart;
                            LoadBuffer(PWAVEHDR(lParam)^.dwUser);
                            waveOutWrite(wParam, PWAVEHDR(lParam), WaveHdrSize);
                            dec(FLoopsToPlay);
                            if FLoopsToPlay<0 then FLoopsToPlay:=0;
                          end else
                            if FPlayingBufferCount<1 then
                              begin
                                //No more data to play => properly end the playback
                                FState:=wpStopped;
                                waveOutReset(wParam); //it may hang up without this
                                waveOutClose(wParam);//Some devices are able to play
                                                 //only one sound=>close it
                                                 //and hope that only one sound
                                                 //is desired to be played
                                for i:=0 to 7 do
                                  freemem(FBuffers[i].lpData);

                                //Call the event as last to ensure that we've
                                //done everything in case that event rises
                                //some exception
                                if Assigned(FOnFinishedPlayback) then
                                  FOnFinishedPlayback(@Self);
                             end;
                      end else
                        waveOutWrite(wParam, PWAVEHDR(lParam), WaveHdrSize);
                 end;
  end;
end;

procedure TWavePlayer.LoadBuffer(AIndex:integer);
var LoadableBufferSize:integer;
begin
  with FBuffers[AIndex], FSource^ do
    begin
      dwFlags:=0;

      LoadableBufferSize:=FDataEnd-Position-FBufferSize;
      if LoadableBufferSize>=0 then LoadableBufferSize:=FBufferSize
        else inc(LoadableBufferSize, FBufferSize);

      dwBufferLength:=Read(lpData^, LoadableBufferSize);
      if dwBufferLength = 0 then exit; //no data to be played

      if waveOutPrepareHeader(FDevice, @FBuffers[AIndex], WaveHdrSize) =
         MMSYSERR_NOERROR then inc(FPlayingBufferCount);
    end;
end;

function TWavePlayer.LoadWaveFile:PWaveFormatEx;
var WaveHdr:TWAVHeader;
    chunk:TChunk;
begin
  result:=nil;
  try
    FSource.Position:=0;
    FSource.Read(WaveHdr, sizeof(TWAVHeader));

    //OK, the we got the header now test it
    if WaveHdr.WAVEHeader <> 'WAVE' then Abort;

    //We're gonna process wave file

    //check for PCM files
    with WaveHdr.FormatEx do
      if wFormatTag = WAVE_FORMAT_PCM then
         begin
           cbSize:=0; //sanity assignment
           FSource.Position:=WaveHdr.FormatSize+20;
                //Some programs takes cbSize as a header member and some don't,
                //but all programs (I've seen) fills FormatSize correctly
         end;

    getmem(result, WaveHdr.FormatEx.cbSize+sizeof(TWaveFormatEx));
    result^:=WaveHdr.FormatEx;

    //And load the rest of format dependent data if any
    FSource.Read(pointer(Cardinal(result)+sizeof(TWaveFormatEx))^,
                       result^.cbSize);

    //Skip all chunks, that we're not interested in
    FSource.Read(chunk, sizeof(chunk));
    while chunk.Signature <> $61746164 do //<> 'data'
      begin
        FSource.position:=FSource.position+chunk.Length; //seek doesn't work
        FSource.Read(chunk, sizeof(chunk));
      end;

    //Remember the file island of the 'data' chunk
    FDataStart:=FSource.position;
    FDataEnd:=FDataStart+chunk.Length;
  except
    if result<>nil then Dispose(Result);
  end;
end;

function TWavePlayer.LoadMPEG1File:PWaveFormatEx;

const
  bufSize = 4096;
  SamplingFreq:array[0..3] of integer = (44100, 48000, 32000, 0);

  ACM_Channels:array[0..3] of integer = (ACM_MPEG_STEREO,
                                         ACM_MPEG_JOINTSTEREO,
                                         ACM_MPEG_DUALCHANNEL,
                                         ACM_MPEG_SINGLECHANNEL);

  ACM_Layers:array[1..3] of integer = (ACM_MPEG_LAYER1,
                                       ACM_MPEG_LAYER2,
                                       ACM_MPEG_LAYER3);

  bitrates:array[0..2, 0..15] of integer = (
                    (0,32,64,96,128,160,192,224,256,288,320,352,384,416,448, 0),
                    (0,32,48,56,64,80,96,112,128,160,192,224,256,320,384, 0),
                    (0,32,40,48,56,64,80,96,112,128,160,192,224,256,320, 0));
        //Note that the bad combination will produce 0 'cause we've scanned
        //only one frame



Type
  PByteArray = ^TByteArray;
  TByteArray = array[0..0] of Byte;
var i, j:integer;
    buf:PByteArray;
    frame:DWORD;       //Sure, it's only the header
    isLayer3:boolean;

begin
  result:=nil;
  frame:=0;
  buf:=nil;
  try
    //Search for synchronization flag of first frame
    FSource.Seek(0,spBegin);
    try
      getmem(buf, bufSize);
      repeat
        FDataStart:=FSource.Position;
        j:=FSource.Read(buf^, bufSize);
        for i:=0 to j-5 do
          if (buf^[i] = $FF) and
             (buf^[i+1] and $F0 = $F0) and
             (buf^[i+1] and 6 <> 0) then
             //OK, it looks like frame we are looking for
             begin
               frame:=PDWORD(longint(buf)+i)^;
               FDataStart:=FDataStart+i;
               j:=0;
               break;
             end;
        if j<>0 then FSource.Seek(-3, spCurrent);
      until j = 0;
    finally
      if buf<>nil then freemem(buf);
    end;

    //If no frame has been found then it is not MPEG1
    if frame = 0 then Abort;

    //Layer I & II are supported by WXP and higher DShow
    isLayer3:=(frame shr 13) and 3 = 3;

    //Fill the wave info

    //Use more supported way if possible
    if isLayer3 then getmem(result, sizeof(TMPEGLAYER3WAVEFORMAT))
      else getmem(result, sizeof(TMPEG1WAVEFORMAT));
    with result^ do
      begin
        if isLayer3 then wFormatTag:=WAVE_FORMAT_MPEGLAYER3
          else wFormatTag:=WAVE_FORMAT_MPEG;
           //3 means mono, else it is stereo
        if (frame shr 30) and 3 = 3 then nChannels:=1
          else nChannels:=2;
        nSamplesPerSec:=SamplingFreq[(frame shr 18) and 3];

        nAvgBytesPerSec:=bitrates[((frame shr 13) and 3)-1,
                                  (frame shr 20) and $f]*1000 shr 3;
        nBlockAlign:=1;
        wBitsPerSample:=0;
        if isLayer3 then cbSize:=sizeof(TMPEGLAYER3WAVEFORMAT)-sizeof(TWAVEFORMATEX)
          else cbSize:=sizeof(TMPEG1WAVEFORMAT)-sizeof(TWAVEFORMATEX);
      end;

      if isLayer3 then
        with PMPEGLAYER3WAVEFORMAT(result)^ do
          begin
            wID:=MPEGLAYER3_ID_MPEG;

            if frame and $20000 <> 0 then fdwFlags:=MPEGLAYER3_FLAG_PADDING_ON
              else fdwFlags:=MPEGLAYER3_FLAG_PADDING_OFF;

            //We have to take care about frequency to be accurate
            with wfx do
              if nSamplesPerSec>=32000 then
                nBlockSize:=1152*nAvgBytesPerSec div nSamplesPerSec
                else nBlockSize:=576*nAvgBytesPerSec div nSamplesPerSec;

            nFramesPerBlock:=1;
            nCodecDelay:=$0571;//If you're interested in this value then search
                               //e.g. for FHG in LAME's sources.
          end else
            with PMPEG1WAVEFORMAT(result)^ do
              begin
                fwHeadLayer:=ACM_Layers[(frame shr 13) and 3];
                dwHeadBitrate:=bitrates[((frame shr 13) and 3)-1,
                                        (frame shr 20) and $f];
                fwHeadMode:=ACM_Channels[(frame shr 30) and 3];
                fwHeadModeExt:=0;     //For encoding and we're player only
                wHeadEmphasis:=((frame shr 24) and 3)+1;
                dwPTSLow:=0;
                dwPTSHigh:=0;
                fwHeadFlags:=ACM_MPEG_ID_MPEG1;
              end;
    FDataEnd:=FSource.Size;
    //MPEG uses synchronization bits, so we can "safely" pass end of source
  except
    if result<>nil then Dispose(result);
  end;
end;


function TWavePlayer.Play:boolean;
var WaveInfo:PWaveFormatEx;
    i:integer;
    res:MMRESULT;

begin
  //There can be a case instead of this 2 ifs, but the compiler will optimize it
  if FState = wpPlaying then
    begin
      result:=false;
      exit;
    end;
  //Check for unpause
  if FState = wpPaused then
    begin
      result:=Pause;
      exit;
    end;
  //if anything will go wrong, then an exception will arise
  WaveInfo:=nil; //keeps compiler quiet
  try
    WaveInfo:=LoadWaveFile;
    if WaveInfo = nil then WaveInfo:=LoadMPEG1File;
    if WaveInfo = nil then Abort;

    FLoopsToPlay:=FDesiredLoops-1; //One loop will be always played
    FPlayingBufferCount:=0;        //None is actually being played

    //Allocate Buffers
    with WaveInfo^ do
      FBufferSize:=nBlockAlign*nSamplesPerSec shr 6; //Like div 64
                //BlocksInBuffer;
    for i:=0 to 7 do
      begin
        getmem(FBuffers[i].lpData, FBufferSize);
        FBuffers[i].dwLoops:=1;
        FBuffers[i].dwUser:=i;
      end;

    //Open the WAVE_MAPPER,
    res:=waveOutOpen(@FDevice, WAVE_MAPPER, WaveInfo,
                     FWnd,
                     cardinal(@Self),
                     CALLBACK_WINDOW or WAVE_ALLOWSYNC);
    if res <> MMSYSERR_NOERROR then
        raise EWPlayer.Create(e_Custom, 'MM System Error: '+int2str(res));
    freemem(WaveInfo);

    //Seek the first data byte
    FSource.Position:=FDataStart;
    //load buffers
    for i:=0 to 7 do
      LoadBuffer(i);

    //and play them.
    FState:=wpPlaying; //Mark desired state
    for i:=0 to 7 do
      if FBuffers[i].dwBufferLength <> 0 then
         waveOutWrite(FDevice, @FBuffers[i], WaveHdrSize);

    result:=true;
  except
    FState:=wpStopped;
    freemem(WaveInfo);
    for i:=0 to 7 do
      freemem(FBuffers[i].lpData);
    result:=false;
    if FBeepAtLeast then Windows.MessageBeep(MB_ICONQUESTION);
    FState:=wpStopped;
  end;
end;

function TWavePlayer.Stop:boolean;
var i:integer;
begin
  result:=false;

  if FState = wpStopped then exit;

  try
    //if not Reset then Abort;  //This call sets FState to wpStopped
    FState:=wpStopped; //This will stop loading the buffers

    //Actually a call to waveOutReset should be here, but it always hangs up
    //current thread => try to close the device right now

    if waveOutClose(FDevice) <> MMSYSERR_NOERROR then
      //Successfull waveOutClose sends WOM_DONE for all buffers
      //=>waveOutProc unprepares and frees them
      begin
        //OK, we can still try to stop the playback without any reset

        for i:=0 to 7 do
          FBuffers[i].dwBufferLength:=0;

        i:=16;   //Countdown - Forever Is Long Long Time
        while (FPlayingBufferCount>0) and (i>0) do
          begin
            Sleep(100);
            dec(i);
          end;

        if waveOutClose(FDevice) <> MMSYSERR_NOERROR then Abort;

        //Because we've waited with demand to immediately stop loading buffers,
        //we've to release them manualy
        for i:=0 to 7 do
          freemem(FBuffers[i].lpData);
      end;

    result:=true;
  except
  end;
end;

function TWavePlayer.Reset:boolean;
begin
  try
    FState:=wpStopped; //--This will at least stop loading the buffers
    if waveOutReset(FDevice) <> MMSYSERR_NOERROR then Abort;

    result:=true;
  except
    result:=false;
  end;
end;

function TWavePlayer.Pause:boolean;
begin
  try
    case FState of
      wpStopped:Abort;
      wpPlaying:begin
                  FState:=wpPaused; //Mark desired state
                  if waveOutPause(FDevice) <> MMSYSERR_NOERROR then Abort;
                end;
      wpPaused: begin
                  FState:=wpPlaying; //Mark desired state
                  if waveOutRestart(FDevice) <> MMSYSERR_NOERROR then Abort;
                end;
    end;
    result:=true;
  except
    result:=false;
  end;
end;

end.
